#    Copyright 2013, 2014, 2015 Joshua Charles Campbell
#
#    This file is part of EstimateCharm.
#
#    EstimateCharm is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EstimateCharm is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with EstimateCharm.  If not, see <http://www.gnu.org/licenses/>.

"""
Typically ran with FAST="True" py.test

See `source_this.sh` in the repository root for enivornment variables that
must be set.
"""

import unittest
from logging import debug, info, warning, error

from estimatecharm.unnaturalCode import *
from estimatecharm.pythonSource import *
from estimatecharm.estimateCharm import *

import os, os.path, zmq, sys, shutil, token, gc
from glob import glob
from tempfile import *

from estimatecharm.ucTestData import *

logging.getLogger(__name__).setLevel(logging.DEBUG)

class testEstimateCharm(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        #self.uc = unnaturalCode()
        pass
    @classmethod
    def tearDownClass(self):
        #del self.uc
        pass

# rwfubmqqoiigevcdefhmidzavjwg



#!/usr/bin/python
#    Copyright 2013, 2014, 2015 Joshua Charles Campbell
#
#    This file is part of EstimateCharm.
#    
#    EstimateCharm is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EstimateCharm is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with EstimateCharm.  If not, see <http://www.gnu.org/licenses/>.

from estimatecharm.unnaturalCode import *
from estimatecharm.pythonSource import *

from logging import debug, info, warning, error
import logging
from random import randint
from os import path
import argparse

import csv
import runpy
import sys, traceback
from shutil import copyfile
from tempfile import mkstemp, mkdtemp
import os, re

from multiprocessing import Process, Queue
try:
  from Queue import Empty
except ImportError:
  from queue import Empty
from estimatecharm import flexibleTokenize

import pdb
import math

virtualEnvActivate = os.getenv("VIRTUALENV_ACTIVATE", None)

nonWord = re.compile('\\W+')
beginsWithWhitespace = re.compile('^\\w')
numeric = re.compile('[0-9]')
punct = re.compile('[~!@#$%^%&*(){}<>.,;\\[\\]`/\\\=\\-+]')
funny = re.compile(flexibleTokenize.Funny)
name = re.compile(flexibleTokenize.Name)

class HaltingError(Exception):
  def __init__(self, value):
    self.value = value
  def __str__(self):
    return repr(self.value)

def runFile(q,path):
    if not virtualEnvActivate is None:
      if sys.version_info >= (3,0):
        exec(compile(open(virtualEnvActivate, "rb").read(), virtualEnvActivate, 'exec'), dict(__file__=virtualEnvActivate))
      else:
        execfile(virtualEnvActivate, dict(__file__=virtualEnvActivate))
    try:
        runpy.run_path(path)
    except SyntaxError as se:
        ei = sys.exc_info();
        eip = (ei[0], str(ei[1]), traceback.extract_tb(ei[2]))
        try:
          eip[2].append(ei[1][1])
        except IndexError:
          eip[2].append((se.filename, se.lineno, None, None))
        q.put(eip)
        return
    except Exception as e:
        ei = sys.exc_info();
        info("run_path exception:", exc_info=ei)
        eip = (ei[0], str(ei[1]), traceback.extract_tb(ei[2]))
        q.put(eip)
        return
    q.put((None, "None", [(path, None, None, None)]))
    
class charmFile(object):
    
    def __init__(self, path, language, tempDir):
        self.path = path
        self.lm = language
        self.f = open(path)
        self.original = self.f.read()
        self.lexed = self.lm(self.original)
        self.scrubbed = self.lexed.scrubbed()
        self.lines = self.lexed[-1].end.line
        self.lineStart = [-1 for i in range(0, self.lines+1)]
        self.lineTokens = [0 for i in range(0, self.lines+1)]
        for i in range(0, len(self.scrubbed)):
          line = self.scrubbed[i].start.line
          self.lineTokens[line] = self.lineTokens[line] + 1
          for j in range(line, 0, -1):
            if self.lineStart[j] == -1:
              self.lineStart[j] = i
            else:
              break
        self.f.close()
        self.mutatedLexemes = None
        self.mutatedLocation = None
        self.tempDir = tempDir
        r = self.run(path)
        info("Ran %s, got %s" % (self.path, r[1]))
        if (r[0] != None):
          raise Exception("Couldn't run file: %s because %s" % (self.path, r[1]))
        #runpy.run_path(self.path)
    
    def run(self, path):
        q = Queue()
        p = Process(target=runFile, args=(q,path,))
        p.start()
        try:
          r = q.get(True, 10)
        except Empty as e:
          r = (HaltingError, "Didn't halt.", [(path, None, None, None)])
        p.terminate()
        p.join()
        assert not p.is_alive()
        #assert r[2][-1][2] != "_get_code_from_file" # This seems to be legit
        return r

    
    def mutate(self, lexemes, location):
        assert isinstance(lexemes, ucSource)
        self.mutatedLexemes = self.lm(lexemes.deLex())
        self.mutatedLocation = location
        
    def runMutant(self):
        (mutantFileHandle, mutantFilePath) = mkstemp(suffix=".py", prefix="mutant", dir=self.tempDir)
        self.mutantFilePath = mutantFilePath
        mutantFile = os.fdopen(mutantFileHandle, "w")
        mutantFile.write(self.mutatedLexemes.deLex())
        mutantFile.close()
        r = self.run(mutantFilePath)
        os.remove(mutantFilePath)
        return r
        
class estimateCharm(object):
    
    def addCharmFile(self, files):
          """Add a file for validation..."""
          files = [files] if isinstance(files, str) else files
          assert isinstance(files, list)
          for fi in files:
            vfi = charmFile(fi, self.lm, self.tempDir)
            if len(vfi.lexed) > 1:
              self.charmFiles.append(vfi)
    
    def estimate(self, mutation, deltamax):
        """Run main estimation loop."""
        for fi in self.charmFiles:
          assert isinstance(fi, charmFile)
          n = len(fi.scrubbed)
          l = fi.lexed[-1].end.line
          if fi.path in self.progress:
            progress = self.progress[fi.path]
            errors = self.errors[fi.path]
            mutations = self.mutations[fi.path]
            charm = self.charm[fi.path]
          else:
            progress = [0 for i in range(1,l+3)]
            progress[0] = None # Line numbers start with 1
            errors = [0 for i in range(1,l+3)]
            errors[0] = None
            charm = [0 for i in range(1,l+3)]
            charm[0] = None
            mutations = 0
          info("Testing " + str(progress) + " " + fi.path)
          delta = float("inf")
          mi = 0
          while (delta > deltamax):
            mi = mi + 1
            mline = (mi % l) + 1
            if fi.lineTokens[mline] > 0:
              merror = mutation(self, fi, mline)
              if merror is not None:
                info(merror)
                break
              runException = fi.runMutant()
              errorLine = None
              filename = None
              func = None
              text = None
              if (runException[0] == None):
                exceptionName = "None"
              else:
                exceptionName = runException[0].__name__
                for location in reversed(runException[2]):
                  if (location[0] == fi.mutantFilePath):
                    filename, errorLine, func, text = location
                    break
              if errorLine == None:
                errorLine = l+1
              if errorLine > l+1: # This can be caused by inserting giant multi-line string literals, in python docstrinsg
                errorLine = l+1
              mutLine = fi.mutatedLocation.start.line
              #info(" ".join(map(str, [fi.path, mutLine, l, fi.mutatedLocation])))
              #info(" ".join(map(str, [filename, errorLine, func, text])))
              #info(runException)
              if (mutLine == errorLine):
                online = True
              else:
                online = False
              errors[errorLine] = errors[errorLine] + 1
              progress[mutLine] = progress[mutLine] + 1
              mutations = mutations + 1
              assert(l>0)
              assert(mutations>0)
              charm[mutLine] = (errors[mutLine]-progress[mutLine])/(float(mutations)/float(l))
              if errorLine <= l:
                charm[errorLine] = (errors[errorLine]-progress[errorLine])/(float(mutations)/float(l))
              delta = 1.0/math.sqrt(float(mutations)/float(l))
              info(" ".join(map(str, [
                  str(mutations) + "/" + str(int(math.ceil(float(l)/(deltamax*deltamax)))),
                  mutLine, errorLine,
                  errors[errorLine],
                  progress[mutLine],
                  charm[mutLine],
                  delta
                ])))
              self.detailsCsv.writerow([
                fi.path, 
                mutLine,
                errorLine,
                errors[errorLine],
                progress[mutLine],
                mutations,
                charm[mutLine],
                delta,
                mutation.__name__, 
                fi.mutatedLocation.type,
                nonWord.sub('', fi.mutatedLocation.value), 
                exceptionName, 
                online,
                filename,
                func])
              self.detailsFile.flush()
          for li in range(1,l):
            self.csv.writerow([
              fi.path,
              li,
              progress[li],
              errors[li],
              charm[li],
              delta
            ])
            
    def deleteRandom(self, vFile):
        """Delete a random token from a file."""
        ls = copy(vFile.scrubbed)
        token = ls.pop(randint(0, len(ls)-1))
        if token.type == 'ENDMARKER':
          return self.deleteRandom(vFile)
        vFile.mutate(ls, token)
        return None
            
    def insertRandom(self, vFile):
        ls = copy(vFile.scrubbed)
        token = ls[randint(0, len(ls)-1)]
        pos = randint(0, len(ls)-1)
        inserted = ls.insert(pos, token)
        if inserted[0].type == 'ENDMARKER':
          return self.insertRandom(vFile)
        vFile.mutate(ls, inserted[0])
        return None
            
    def replaceRandom(self, vFile, targetLine=None):
        ls = copy(vFile.scrubbed)
        token = ls[randint(0, len(ls)-1)]
        if targetLine == None:
          pos = randint(0, len(ls)-2)
        else:
          lineStart = vFile.lineStart[targetLine]
          nextLineStart = len(vFile.scrubbed)
          if targetLine < vFile.lines:
            nextLineStart = vFile.lineStart[targetLine+1]
          if (lineStart > nextLineStart):
            pos = randint(lineStart, nextLineStart-1)
          else:
            pos = lineStart
          #print str(targetLine)
          #print repr(ls[pos])
          assert(ls[pos].start.line <= targetLine and targetLine <= ls[pos].end.line)
        oldToken = ls.pop(pos)
        if oldToken.type == 'ENDMARKER':
          return self.replaceRandom(vFile)
        inserted = ls.insert(pos, token)
        if inserted[0].type == 'ENDMARKER':
          return self.replaceRandom(vFile)
        vFile.mutate(ls, inserted[0])
        return None
        
    def dedentRandom(self, vFile):
        s = copy(vFile.original)
        lines = s.splitlines(True);
        while True:
          line = randint(0, len(lines)-1)
          if beginsWithWhitespace.match(lines[line]):
            lines[line][0] = ''
            break
        vFile.mutatedLexemes = vFile.lm("".join(lines))
        vFile.mutatedLocation = pythonLexeme.fromTuple((token.INDENT, ' ', (line+1, 0), (line+1, 0)))
        return None
        
    def indentRandom(self, vFile):
        s = copy(vFile.original)
        lines = s.splitlines(True);
        line = randint(0, len(lines)-1)
        if beginsWithWhitespace.match(lines[line]):
          lines[line] = lines[line][0] + lines[line]
        else:
          lines[line] = " " + lines[line]
        vFile.mutatedLexemes = vFile.lm("".join(lines))
        vFile.mutatedLocation = pythonLexeme.fromTuple((token.INDENT, ' ', (line+1, 0), (line+1, 0)))
        return None
    
    def punctRandom(self, vFile):
        s = copy(vFile.original)
        charPos = randint(1, len(s)-1)
        linesbefore = s[:charPos].splitlines(True)
        line = len(linesbefore)
        lineChar = len(linesbefore[-1])
        c = s[charPos:charPos+1]
        if (funny.match(c)):
          new = s[:charPos] + s[charPos+1:]
          vFile.mutatedLexemes = vFile.lm(new)
          vFile.mutatedLocation = pythonLexeme.fromTuple((token.OP, c, (line, lineChar), (line, lineChar)))
          return None
        else:
          return self.punctRandom(vFile)
    
    #def keyRandom(self, vFile):
        #s = copy(vFile.original)
        
    def nameRandom(self, vFile):
      return self.deleteWordRandom(vFile)

    def insertWordRandom(self, vFile):
        s = copy(vFile.original)
        while True:
          char = s[randint(1, len(s)-1)]
          charPos = randint(1, len(s)-1)
          linesbefore = s[:charPos].splitlines(True)
          line = len(linesbefore)
          lineChar = len(linesbefore[-1])
          c = s[charPos:charPos+1]
          if (name.match(char)):
            break
        new = s[:charPos] + char + s[charPos:]
        vFile.mutatedLexemes = vFile.lm(new)
        vFile.mutatedLocation = pythonLexeme.fromTuple((token.OP, c, (line, lineChar), (line, lineChar)))
        return None

    def deleteWordRandom(self, vFile):
        s = copy(vFile.original)
        while True:
          charPos = randint(1, len(s)-1)
          linesbefore = s[:charPos].splitlines(True)
          line = len(linesbefore)
          lineChar = len(linesbefore[-1])
          c = s[charPos:charPos+1]
          if (name.match(c)):
            break
        new = s[:charPos] + s[charPos+1:]
        vFile.mutatedLexemes = vFile.lm(new)
        vFile.mutatedLocation = pythonLexeme.fromTuple((token.OP, c, (line, lineChar), (line, lineChar)))
        return None
        
    def insertPunctRandom(self, vFile):
        s = copy(vFile.original)
        if not punct.search(s):
          return "No punctuation"
        while (True):
          char = s[randint(1, len(s)-1)]
          if (punct.match(char)):
            break
        charPos = randint(1, len(s)-1)
        linesbefore = s[:charPos].splitlines(True)
        line = len(linesbefore)
        lineChar = len(linesbefore[-1])
        c = s[charPos:charPos+1]
        new = s[:charPos] + char + s[charPos:]
        vFile.mutatedLexemes = vFile.lm(new)
        vFile.mutatedLocation = pythonLexeme.fromTuple((token.OP, c, (line, lineChar), (line, lineChar)))
        return None

    def deleteNumRandom(self, vFile):
        s = copy(vFile.original)
        if not numeric.search(s):
          return "No numbers"
        positions = [x.start() for x in numeric.finditer(s)]
        while True:
          if (len(positions) == 1):
            charPos = positions[0]
          else:
            charPos = positions[randint(1, len(positions)-1)]
          linesbefore = s[:charPos].splitlines(True)
          line = len(linesbefore)
          lineChar = len(linesbefore[-1])
          c = s[charPos:charPos+1]
          if (numeric.match(c)):
            break
        new = s[:charPos] + s[charPos+1:]
        vFile.mutatedLexemes = vFile.lm(new)
        vFile.mutatedLocation = pythonLexeme.fromTuple((token.OP, c, (line, lineChar), (line, lineChar)))
        return None

    def insertNumRandom(self, vFile):
        s = copy(vFile.original)
        char = str(randint(0, 9))
        charPos = randint(1, len(s)-1)
        linesbefore = s[:charPos].splitlines(True)
        line = len(linesbefore)
        lineChar = len(linesbefore[-1])
        c = s[charPos:charPos+1]
        new = s[:charPos] + char + s[charPos:]
        vFile.mutatedLexemes = vFile.lm(new)
        vFile.mutatedLocation = pythonLexeme.fromTuple((token.OP, c, (line, lineChar), (line, lineChar)))
        return None

    def deletePunctRandom(self, vFile):
        s = copy(vFile.original)
        if not punct.search(s):
          return "No punctuation"
        while True:
          charPos = randint(1, len(s)-1)
          linesbefore = s[:charPos].splitlines(True)
          line = len(linesbefore)
          lineChar = len(linesbefore[-1])
          c = s[charPos:charPos+1]
          if (punct.match(c)):
            break
        new = s[:charPos] + s[charPos+1:]
        vFile.mutatedLexemes = vFile.lm(new)
        vFile.mutatedLocation = pythonLexeme.fromTuple((token.OP, c, (line, lineChar), (line, lineChar)))
        return None

    def colonRandom(self, vFile):
        s = copy(vFile.original)
        while True:
          charPos = randint(1, len(s)-1)
          linesbefore = s[:charPos].splitlines(True)
          line = len(linesbefore)
          lineChar = len(linesbefore[-1])
          c = s[charPos:charPos+1]
          if (c == ':'):
            break
        new = s[:charPos] + s[charPos+1:]
        vFile.mutatedLexemes = vFile.lm(new)
        vFile.mutatedLocation = pythonLexeme.fromTuple((token.OP, c, (line, lineChar), (line, lineChar)))
        return None
      
    def __init__(self, source=None,
                 language=pythonSource,
                 results=None,
                 corpus=None,
                 details=None,
                 activate=None,
                 tempDir="."):
        if isinstance(source, str):
            raise NotImplementedError
        elif isinstance(source, list):
            self.charmFileNames = source
        else:
            raise TypeError("Constructor arguments!")
        self.notReleased = True
        self.progress = dict()
        self.results = results
        self.details = details
        self.tempDir = tempDir
        try:
          self.csvFile = open(self.results, 'r')
          self.csv = csv.reader(self.csvFile)
          for row in self.csv:
              self.progress[row[0]][row[1]] = row[2]
              self.mutations[row[0]] = self.mutations[row[0]] + row[2]
              self.errors[row[0]][row[1]] = row[3]
              self.charm[row[0]][row[1]] = row[4]
          self.csvFile.close()
        except (IOError):
          pass
        self.csvFile = open(self.results + ".new", 'w')
        self.csv = csv.writer(self.csvFile)
        self.csv.writerow([
          "file",
          "line",
          "mutants",
          "errors",
          "charm",
          "delta"
        ])
        self.detailsFile = open(self.details, 'a')
        self.detailsCsv = csv.writer(self.detailsFile)
        self.lm = language
        self.charmFiles = list()
        self.addCharmFile(self.charmFileNames)

    def release(self):
        self.notReleased = False
        """Any cleanup goes here..."""
        
    def __del__(self):
        """I am a destructor, but release should be called explictly."""
        assert not self.notReleased, "Destructor called before release()"

DELETE = estimateCharm.deleteRandom
INSERT = estimateCharm.insertRandom
REPLACE = estimateCharm.replaceRandom
PUNCTUATION = estimateCharm.punctRandom
NAMELIKE = estimateCharm.nameRandom
COLON = estimateCharm.colonRandom
DELETEWORDCHAR = estimateCharm.deleteWordRandom
INSERTWORDCHAR = estimateCharm.insertWordRandom
DELETENUMCHAR = estimateCharm.deleteNumRandom
INSERTNUMCHAR = estimateCharm.insertNumRandom
DELETEPUNCTCHAR = estimateCharm.deletePunctRandom
INSERTPUNCTCHAR = estimateCharm.insertPunctRandom
DELETESPACE = estimateCharm.dedentRandom
INSERTSPACE = estimateCharm.indentRandom

def main():
        logging.getLogger().setLevel(logging.DEBUG)
        parser=argparse.ArgumentParser(description="Estimates charm for Python source code.")
        parser.add_argument("input_file", help="Python source file to estimate charm for.", nargs="+")
        parser.add_argument("-o", "--results-file", help="File to store results in.", default="charm.csv")
        parser.add_argument("-d", "--details-file", help="File to store extra detailed results in.", default="detail.csv")
        parser.add_argument("-a", "--activate", help="VirtualEnv activate.py to run before input files (if any)", default=None)
        parser.add_argument("-e", "--maximum-error", help="Sets the maximum allowed error (the minimum precision) of the results", default=0.1, type=float)
        args = parser.parse_args()
        v = estimateCharm(source=args.input_file, 
                          language=pythonSource,
                          results=args.results_file,
                          details=args.details_file,
                          activate=args.activate
                         )
        v.estimate(REPLACE, args.maximum_error)
        v.release()

if __name__ == '__main__':
    main()



"""

"""

import unittest
from logging import debug, info, warning, error

from estimatecharm.unnaturalCode import *
from estimatecharm.pythonSource import *
from estimatecharm.estimateCharm import *

import os, os.path, sys, shutil, token, gc
from tempfile import *

from estimatecharm.ucTestData import *

# Helper. Given a list of lexemes, gives the stringified version
# of the ith lexeme.
strlex = lambda lexemes, i: str(pythonSource(lexemes)[i])
# Helper. Same as strlex, but takes a single dict representing a lexeme.
strlexd = lambda lexeme: str(pythonLexeme.fromDict(lexeme))

logging.getLogger(__name__).setLevel(logging.DEBUG)


class testPythonLexical(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        pass
    def testLexExpectedLength(self):
        r = pythonSource(somePythonCode)
        self.assertEquals(len(r), 9)
    def testLexExpectedFormat(self):
        r = pythonSource(somePythonCode)
        self.assertTrue(isinstance(r, ucSource))
        self.assertTrue(isinstance(r[0], ucLexeme))
        self.assertTrue(isinstance(r[0].end, tuple))
        self.assertTrue(isinstance(r[0].end.l, int))
        self.assertTrue(isinstance(r[0].end.c, int))
        self.assertTrue(isinstance(r[0].start, tuple))
        self.assertTrue(isinstance(r[0].start[0], int))
        self.assertTrue(isinstance(r[0].start[1], int))
        self.assertTrue(isinstance(r[0].type, str))
        self.assertTrue(isinstance(r[0].value, str))
    def testLexExpectedToken(self):
        r = pythonSource(somePythonCode)
        self.assertEquals(r[0].end[0], 1)
        self.assertEquals(r[0].end[1], 5)
        self.assertEquals(r[0].start[0], 1)
        self.assertEquals(r[0].start[1], 0)
        self.assertEquals(r[0].type, 'NAME')
        self.assertEquals(r[0].value, 'print')
    def testLexMidlineModeSingleLine(self):
        r_normal = pythonSource(somePythonCode)
        r_mid_line = pythonSource(somePythonCode, mid_line=True)
        # This small one line example has no indent or anything.
        self.assertEquals(len(r_normal), 9)
        self.assertEquals(len(r_mid_line), 8)
        self.assertEquals(r_normal[-1].type, 'ENDMARKER')
        self.assertNotEquals(r_mid_line[-1].type, 'ENDMARKER')
    def testLexMidlineModeExcerpt(self):
        r_normal = pythonSource(incompletePythonCode)
        r_mid_line = pythonSource(incompletePythonCode, mid_line=True)
        self.assertEquals(len(r_normal), 13)
        self.assertEquals(len(r_mid_line), 11)
        self.assertEquals(r_normal[-1].type, 'ENDMARKER')
        self.assertEquals(r_normal[-2].type, 'DEDENT')
        self.assertEquals(r_mid_line[-1].value, 'ran')
    def testColumns(self):
        r = pythonSource(lotsOfPythonCode)
        self.assertEquals(r[1].columns(), 3) # this should be the "def" token
        self.assertEquals(r[0].lines(), 0)
    def testDeleteOne(self):
        r = pythonSource(lotsOfPythonCode)
        x = r.pop(1)
        self.assertEquals(x.value, 'def')
        r.check()
        r.pop(6)
        r.check()
        r.pop(1)
        r.check()
        r.pop(1)
        r.check()
        r.pop(1)
        r.check()
        r.pop(1)
        r.check()
        r.pop(1)
        r.check()
        r.pop(1)
        r.check()
        r.pop(1)
        r.check()
        r.pop(1)
        r.check()
        r.pop(1)
        r.check()
        r.pop(1)
        r.check()
    def testDeleteOneTwo(self):
        r = pythonSource(codeWithDeleteFailure)
        x = r.pop(2)
        r.check()
        r.pop(2)
        r.check()
        r.pop(2)
        r.check()
    def testStringifyStandard(self):
        self.assertEquals(str(pythonSource(someLexemes)[0]), 'print')
        self.assertEquals(str(pythonSource(someLexemes)[8]), '<ENDMARKER>')
        self.assertEquals(str(pythonLexeme.fromDict(indentLexeme)), '<INDENT>')
    def testStringifyLegacyBehaviour(self):
        tooLong = {'type': 'NAME', 'start': (0, 0), 'end': (0, 45),
                   'value': 'iAmADickBagIdentifierWithMoreThan32Characters'}
        comment = {'type': 'COMMENT', 'start': (0, 0), 'end': (0, 35),
                   'value': '#!/bin/dd if=/dev/zero of=/dev/sda1'}
        string = {'type': 'STRING', 'start': (0, 0), 'end': (0, 35),
                  'value': 'u"Hello, World!"'}
        self.assertEquals(strlexd(tooLong), '<NAME>')
        self.assertEquals(strlexd(comment), '<COMMENT>')
        self.assertEquals(strlexd(string), '<STRING>')
    def testStringifyDedent(self):
        self.assertEquals(strlex([dedentLexeme], 0), '<DEDENT>')
        self.assertEquals(str(pythonLexeme.fromDict(dedentLexeme)), '<DEDENT>')
    def testLexDeLex(self):
        self.assertEquals(lotsOfPythonCode, (pythonSource(lotsOfPythonCode).deLex()))
        self.assertEquals(codeWithComments, (pythonSource(codeWithComments).deLex()))
    def testComment(self):
        self.assertTrue(pythonLexeme.fromTuple((COMMENT, '# wevie stunder', (1, 0), (1, 0))).comment())
        self.assertFalse(pythonLexeme.fromTuple((token.INDENT, '    ', (2, 0), (2, 0))).comment())
    def testExclusiveEnd(self):
        r = pythonSource("1+2")
        self.assertTrue(r[0].start.c + 1 == r[0].end.c) # First token is r[0]
        self.assertTrue(r[2].start.c + 1 == r[2].end.c)
    def testPoppin(self):
        r = pythonSource("1+2")
        r.pop(0)
        r.pop(0)
        r.pop(0)
    def testPopBug(self):
        # AssertionError: [('OP', '=', (76, 44), (76, 45)), ('NEWLINE', '\n', (76, 3), (76, 4))]
        r = pythonSource("a=1+2\na")
        r.pop(4)
        r.check()
        r = pythonSource("a=1+2\na")
        r.pop(5)
        r.check()
        r = pythonSource("a=1+2\n\n")
        r.pop(4)
        r.check()
        r = pythonSource("a=1\n\n")
        r.pop(2)
        r.check()
        r = pythonSource("for _ in range(0, x):\n\ta=1\n\n")
        x = r.pop(14)
        r.check()
        self.assertEquals(x.value, '1')
        r = pythonSource("for _ in range(0, x):\n\ta=1\n\n")
        x = r.pop(9)
        r.check()
        self.assertEquals(x.value, ':')




#    Copyright 2013, 2014, 2015 Joshua Charles Campbell
#
#    This file is part of EstimateCharm.
#    
#    EstimateCharm is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EstimateCharm is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with EstimateCharm.  If not, see <http://www.gnu.org/licenses/>.
import os
from pkg_resources import resource_filename
import fnmatch

somePythonCode = "print (1+2**2)"

incompletePythonCode = """def foo():
    for i in ran"""

someLexemes = [{'end': (1, 5), 'start': (1, 0), 'type': 'NAME', 'value': 'print'},
                {'end': (1, 8), 'start': (1, 7), 'type': 'OP', 'value': '('},
                {'end': (1, 9), 'start': (1, 8), 'type': 'NUMBER', 'value': '1'},
                {'end': (1, 10), 'start': (1, 9), 'type': 'OP', 'value': '+'},
                {'end': (1, 11), 'start': (1, 10), 'type': 'NUMBER', 'value': '2'},
                {'end': (1, 13), 'start': (1, 11), 'type': 'OP', 'value': '**'},
                {'end': (1, 14), 'start': (1, 13), 'type': 'NUMBER', 'value': '2'},
                {'end': (1, 15), 'start': (1, 14), 'type': 'OP', 'value': ')'},
                {'end': (2, 0), 'start': (2, 0), 'type': 'ENDMARKER', 'value': ''}]

indentLexeme =  {'end': (3, 8), 'start': (3, 0), 'type': 'INDENT', 'value': '        '}
dedentLexeme =  {'start': (3, 0), 'end': (3, 0), 'type': 'DEDENT', 'value': ''}

lotsOfPythonCode = """
def mult(x, y):
    r = 0
    for _ in range(0, x):
        r = r + y

print mul(1, 2)


"""

codeWithComments = """
# this is the multiplcation algorithm
def mult(x, y):
    # initialize
    r = 0
    for _ in range(0, x): # loop and stuff
        r = r + y # inside of the loop

print mul(1, 2) # output the result
"""

codeWithDeleteFailure = """
null_translationimportqueueentry_potemplate = \"\"\"\\
UPDATE TranslationImportQueueEntry
   SET potemplate = NULL
 WHERE TranslationImportQueueEntry.id IN (
    SELECT TranslationImportQueueEntry.id
      FROM TranslationImportQueueEntry, POTemplate
     WHERE TranslationImportQueueEntry.potemplate = POTemplate.id
       AND POTemplate.distroseries = ?
     LIMIT ?)
\"\"\"
"""

testFileList = os.getenv("TEST_FILE_LIST", None)

testProjectFiles = []

if testFileList == None:
  testProjectFiles = []
else:          
  testProjectFiles = open(testFileList).read().splitlines()
  
  
# DEAR EDDIE, THIS IS THE CODE YOU SHOULD TRY TO USE TO GET YOUR 
# INPUT CORPUS FOR YOUR TESTS
if len(testProjectFiles) < 11:
  testDataDir = resource_filename('unnaturalcode', "testdata")
  for dirPath, dirNames, fileNames in os.walk(testDataDir):
    for pythonFile in fnmatch.filter(fileNames, '*.py'):
      testProjectFiles.append(os.path.abspath(os.path.join(dirPath, pythonFile)))

if len(testProjectFiles) < 11:
  raise RuntimeError("Not enough testing files! ")

testProject1File = testProjectFiles[10]

somePythonCodeFromProject = """
submission_set = getUtility(IHWSubmissionSet)
submission_2 = submission_set.getBySubmissionKey('submission-2')
self.assertEqual(
    submission_2.status, HWSubmissionProcessingStatus.SUBMITTED,
    'Unexpected status of submission 1: %s' % submission_2.status)
"""




#!/usr/bin/python
#    Copyright 2013, 2014, 2015 Joshua Charles Campbell, Alex Wilson
#
#    This file is part of EstimateCharm.
#    
#    EstimateCharm is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EstimateCharm is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with EstimateCharm.  If not, see <http://www.gnu.org/licenses/>.

from estimatecharm.unnaturalCode import *
from logging import debug, info, warning, error

from estimatecharm import flexibleTokenize
import re

import sys, token
try:
  from cStringIO import StringIO
except ImportError:
  from io import StringIO

COMMENT = 53

ws = re.compile('\s')

class pythonLexeme(ucLexeme):
    
    @classmethod
    def stringify(cls, t, v):
        """
        Stringify a lexeme: produce a string describing it.
        In the case of comments, strings, indents, dedents, and newlines, and
        the endmarker, a string with '<CATEGORY-NAME>' is returned.  Else, its
        actual text is returned.
        """
        if t == 'COMMENT':
            return '<'+t+'>'
        # Josh though this would be a good idea for some strange reason:
        elif len(v) > 20 :
            return '<'+t+'>'
        elif ws.match(str(v)) :
            return '<'+t+'>'
        elif t == 'STRING' :
            return '<'+t+'>'
        elif len(v) > 0 :
            return v
        else:
            # This covers for <DEDENT> case, and also, probably some other
            # special cases...
            return '<' + t + '>'
    
    @classmethod
    def fromTuple(cls, tup):
        if isinstance(tup[0], int):
            t0 = token.tok_name[tup[0]]
        else:
            t0 = tup[0]
        return tuple.__new__(cls, (t0, str(tup[1]), ucPos(tup[2]), ucPos(tup[3]), cls.stringify(t0, str(tup[1]))))
          
    def comment(self):
        return (self.ltype == 'COMMENT')
      

class pythonSource(ucSource):
    
    def lex(self, code, mid_line=False):
        tokGen = flexibleTokenize.generate_tokens(StringIO(code).readline,
            mid_line)
        return [pythonLexeme.fromTuple(t) for t in tokGen]
    
    def deLex(self):
        line = 1
        col = 0
        src = ""
        for l in self:
            for i in range(line, l.start.line):
                src += os.linesep
                col = 0
                line += 1
            for i in range(col, l.start.col):
                src += " "
                col += 1
            src += l.val
            col += len(l.val)
            nls = l.val.count(os.linesep)
            if (nls > 0):
                line += nls
                col = len(l.val.splitlines().pop())
        return src
    
    def unCommented(self):
        assert len(self)
        return filter(lambda a: not a.comment(), copy(self))
    
    def scrubbed(self):
        """Clean up python source code removing extra whitespace tokens and comments"""
        ls = copy(self)
        assert len(ls)
        i = 0
        r = []
        for i in range(0, len(ls)):
            if ls[i].comment():
                continue
            elif ls[i].ltype == 'NL':
                continue
            elif ls[i].ltype == 'NEWLINE' and i < len(ls)-1 and ls[i+1].ltype == 'NEWLINE':
                continue
            elif ls[i].ltype == 'NEWLINE' and i < len(ls)-1 and ls[i+1].ltype == 'INDENT':
                continue
            else:
                r.append(ls[i])
        assert len(r)
        return pythonSource(r)


#!/usr/bin/env python
import sys
from setuptools import setup, find_packages

from estimatecharm import __version__

with open('requirements.txt') as f:
    requires = [l.strip() for l in f.readlines()]

with open('test-requirements.txt') as f:
    tests_require = [l.strip() for l in f.readlines()]

setup(
    name = "estimatecharm",
    version = __version__,
    packages = find_packages(
        #exclude = ["testdata"]
      ),
    entry_points = {
        "console_scripts": [
            "estimatecharm = estimatecharm.estimateCharm:main",
        ],
    },
    author = "Joshua Charles Campbell",
    description = "Code Charm Estimation Tool",
    author_email='joshua2@ualberta.ca',
    url='https://github.com/orezpraw/estimate-charm',
    download_url='https://github.com/orezpraw/estimate-charm/tarball/0.1',
    license='AGPL3+',
    include_package_data = True,
    install_requires = requires,
    tests_require=tests_require,
    zip_safe = False,
)



#    Copyright 2013, 2014, 2015 Joshua Charles Campbell
#
#    This file is part of EstimateCharm.
#    
#    EstimateCharm is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EstimateCharm is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with EstimateCharm.  If not, see <http://www.gnu.org/licenses/>.
import sys, os
import logging
from logging import debug, info, warning, error
from copy import copy

if hasattr(sys, 'maxint'): # Python 2/3 Compatibility
  maxint = sys.maxint
else:
  maxint = sys.maxsize

zctx = None

ucParanoid = os.getenv("PARANOID", False)

class ucPos(tuple):
    def __new__(cls, *args):
        if isinstance(args[0], ucPos):
            return args[0]
        elif len(args) == 2:
            l = args[0]
            c = args[1]
        elif isinstance(args[0], tuple):
            (l, c) = args[0]
        else:
          raise TypeError("Bad constructor arguments.")
        return tuple.__new__(cls, (l, c))
    
    if ucParanoid:
        def __init__(self, *args):
            assert isinstance(self[0], int)
            assert isinstance(self[1], int)
            assert self[0] >= 1
            assert self[1] >= 0
    
    def __getattr__(self, name):
        if name[0] == 'l':
          return self[0]
        elif name[0] == 'c':
          return self[1]
        raise AttributeError
    
    def __str__(self):
        return str(self[0]) + ":" + str(self[1])
    
    def __eq__(self, other):
        return (self[0] == other[0]) and (self[1] == other[1])
    
    def __ne__(self, other):
      return not self.__eq__(other)
    
    def __gt__(self, other):
        return (self[0] > other[0]) or ((self[0] == other[0]) and (self[1] > other[1]))
      
    def __lt__(self, other):
        return not self.__gt__(other)
    
    def __ge__(self, other):
        return self.__gt__(other) or self.__eq__(other)
      
    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)
      
class ucLexeme(tuple):
    if ucParanoid:
        def __init__(self, *args):
            assert len(self) == 5
            assert self[0]
            assert isinstance(self[0], str)
            assert len(self[0] > 0)
            assert isinstance(self[1], str)
            assert len(self[1] > 0)
            assert isinstance(self[2], ucPos)
            assert isinstance(self[3], ucPos)
            assert self[2] <= self[3], "%s > %s" % (self[2], self[3])
            assert isinstance(self[4], str)
            assert len(self[4] > 0)
            
    
    def __getattr__(self, name):
        if name == 'ltype' or name == 'type':
            return self[0]
        elif name == 'val' or name == 'value':
            return self[1]
        elif name == 'start':
            return self[2]
        elif name == 'end':
            return self[3]
        raise AttributeError

    def comment(self):
        return False
    
    def columns(self):
        if self[2][0] == self[3][0]:
            return self[3][1] - self[2][1]
        else:
            return 0
    
    def lines(self):
         return self[3][0] - self[2][0]
    
    @classmethod
    def stringify(cls, t, v):
        if v:
            return v
        else:
            return '<'+t+'>'
        
    @classmethod
    def fromTuple(cls, tup):
        if len(args[0] == 4):
            t = (args[0][0], args[0][1], ucPos(args[0][2]), ucPos(args[0][3]), cls.stringify(args[0][0], args[0][1]))
        elif len(args[0] == 5):
            t = (args[0][0], args[0][1], ucPos(args[0][2]), ucPos(args[0][3]), args[0][4])
        else:
            raise TypeError("Constructor argument cant be " + str(type(args[0])))
        return cls(t)

    
    @classmethod
    def fromDict(cls, d):
        if isinstance(d, dict):
            t = (d['type'], d['value'],  ucPos(d['start']),  ucPos(d['end']), cls.stringify(d['type'], d['value']))
        else:
            raise TypeError("Constructor argument cant be " + str(type(d)))
        return cls(t)
    
    @classmethod
    def build(cls, *args):
        """Initialize a lexeme object."""
        if isinstance(args[0], cls):
            return args[0]
        elif len(args) == 4:
            t = (args[0], args[1], ucPos(args[2]), ucPos(args[3]), cls.stringify(args[0], args[1]))
        elif len(args) == 5:
            t = (args[0], args[1], ucPos(args[2]), ucPos(args[3]), args[4])
        else:
            raise TypeError("Constructor arguments cant be " + str(type(args)))
        return cls(t)
        

    def __str__(self):
        return self[4]

    
class ucSource(list):
    
    def __init__(self, value=[], **kwargs):
        if isinstance(value, str):
            self.extend(self.lex(value, **kwargs))
        elif isinstance(value, list):
            if len(value) == 0:
                return
            elif isinstance(value[0], dict):
                self.extend(map(ucLexeme.fromDict, value))
            else:
                self.extend(map(ucLexeme, value))
        else:
            raise AttributeError

    def settle(self):
        """Contents may settle during shipping."""
        first = ucPos(self[0].start)
        for i in range(0, len(self)):
            if (self[i].end.l - self[i].start.l) > 1:
                endL = self[i].start.l + 1
            else:
                endL = self[i].end.l
            if self[i].start.l == first.l:
                startL = 1
                startC = self[i].start.c - first.c
            else:
                startL = self[i].start.l - first.l-1
                startC = self[i].start.c
            if endL == first.l:
                endL = 1
                endC = self[i].end.c - first.c
            else:
                endL = endL - (first.l-1)
                endC = self[i].end.c
            self[i] = self[i].__class__((self[i][0], self[i][1], ucPos((startL, startC)), ucPos((endL, endC)), self[i][4]))
        if ucParanoid:
            self.check()
        return self
    
    def check(self, start=0, end=maxint):
      start = max(start, 0)
      end = min(end, len(self))
      #debug(str(start) + "-" + str(end))
      for i in range(start, end-1):
        assert isinstance(self[i], ucLexeme)
        assert self[i].end <= self[i+1].start, "In file: %s %s"  % (currentValidationFileForDebuggingPurposesOnly, ""+repr(self[i:i+2]))
    
    if ucParanoid:
        def extend(self, arg):
            for a in arg:
                assert isinstance(a, ucLexeme)
            s = len(self)-1
            r = super(ucSource, self).extend(arg)
            if s >= 0:
                self.check(start=s)
            return r
    
        def append(self, *args):
            return self.extend(args)
      
    def insert(self, i, arg):
        assert i < len(self)
        assert i >= 0
        if not isinstance(arg, list):
          arg = [arg]
        if not isinstance(arg, ucSource):
          arg = ucSource(arg)
        a = copy(arg)
        a.settle()
        width = 0
        if (a[0].start.l == a[-1].end.l):
          width = a[-1].end.c - a[0].start.c
        height = a[-1].end.l - a[0].start.l
        oldend = self[-1].end.l
        for j in range(0, len(a)):
          ((startL, startC), (endL, endC)) = (a[j].start, a[j].end)
          if j == 0:
            startC += 1
          if startL == 1:
            startC += self[i].start.c
          if endL == 1:
            endC += self[i].start.c
          startL += self[i].start.l-1
          endL += self[i].start.l-1
          if j == len(a)-1:
            endC += 1
          a[j] = a[j].__class__((a[j][0], a[j][1], ucPos((startL, startC)), ucPos((endL, endC)), a[j][4]))
        for j in range(i, len(self)):
          ((startL, startC), (endL, endC)) = (self[j].start, self[j].end)
          if startL == a[-1].end.l:
            startC += width
          if endL == a[-1].start.l:
            endC += width
          startL += height
          endL += height
          self[j:j+1] = [self[j].__class__((self[j][0], self[j][1], ucPos((startL, startC)), ucPos((endL, endC)), self[j][4]))]
        for j in range(0, len(a)):
          r = super(ucSource, self).insert(i+j, a[j])
        #if ( self[-1].end.l > oldend +1):
          #info(str(self[-1].end.l) + " > " + str(oldend + 1))
          #assert(false, 'waugh!')
        if ucParanoid:
            self.check()
        return a
    
    def pop(self, i):
        assert i < len(self)
        assert i >= 0
        r = super(ucSource, self).pop(i)
        removedLines = r.lines()
        for j in range(i, len(self)):
          ((startL, startC), (endL, endC)) = (self[j].start, self[j].end)
          if startL == r.end.l:
            startC += r.start.c - r.end.c
            if endL == startL:
              endC += r.start.c - r.end.c
          startL -= r.lines()
          endL -= r.lines()
          self[j:j+1] = [self[j].__class__((self[j][0], self[j][1], ucPos((startL, startC)), ucPos((endL, endC)), self[j][4]))]
        if ucParanoid:
            self.check()
        return r

    def scrubbed(self):
        raise NotImplementedError
        
    if ucParanoid:
        def __setitem__(self, index, value):
            if isinstance(value, list):
                for i in value:
                    assert isinstance(i, ucLexeme)
            else:
                assert isinstance(value, ucLexeme)
            r = super(ucSource, self).__setitem__(index, value)
            if isinstance(index, int):
                self.check(index-1, index+2)
            return r
       
        # Taken from the python documentation: http://docs.python.org/2/reference/datamodel.html v2.7.5 November 2013
        # Licesne: PSF
        def __setslice__(self, i, j, seq):
            self[max(0, i):max(0, j):] = seq
        # End License
    
    def sort():
      raise TypeError("Je refuse!")
    
    
      
# rwfubmqqoiigevcdefhmidzavjwg
